#import "ADOperation+ADTConstructors.h"
#import "ADTMacroses.h"
#import "ADTOperationCounter.h"

#import <AsyncDispatcher/AsyncDispatcher.h>

@interface ADSequenceTest : GHAsyncTestCase

@end

@implementation ADSequenceTest

-(void)testSequenceOfBlocks
{
   [ self prepare ];

   __block ADTOperationCounter recent_counter_ = 0;

   NSRange operations_range_ = NSMakeRange( 0, 15 );

   //{0..14}
   ADSequence* sequence_ = [ ADSequence compositeWithOperations: [ NSArray arrayWithOperationsFromRange: operations_range_
                                                                                              doneBlock: ADT_CHECK_AND_INCREMENT_COUNT( recent_counter_ ) ]
                                                           name: @"global"
                                                      doneBlock: ADT_CHECK_TOTAL_COUNT( recent_counter_, NSMaxRange( operations_range_ ) ) ];

   [ sequence_ async ];

   [ self waitForStatus: kGHUnitWaitStatusSuccess timeout: 2.0 ];
}

-(void)testEmptySequence
{
   [ self prepare ];

   ADSequence* sequence_ = [ ADSequence compositeWithOperations: nil
                                                           name: @"global"
                                                      doneBlock:
                            ADDoneOnMainThread
                            (
                             ^( id< ADResult > result_ )
                             {
                                if ( [ NSThread isMainThread ] )
                                {
                                   [ self notify: kGHUnitWaitStatusSuccess forSelector: _cmd ];
                                }
                             })];

   [ sequence_ async ];

   [ self waitForStatus: kGHUnitWaitStatusSuccess timeout: 0.1 ];
}

-(void)testCopy
{
   [ self prepare ];
   
   __block ADTOperationCounter recent_counter_ = 0;
   
   NSRange operations_range_ = NSMakeRange( 0, 15 );
   
   //{0..14}
   ADSequence* sequence_ = [ ADSequence compositeWithOperations: [ NSArray arrayWithOperationsFromRange: operations_range_
                                                                                              doneBlock: ADT_CHECK_AND_INCREMENT_COUNT( recent_counter_ ) ]
                                                           name: @"global"
                                                      doneBlock: ADT_CHECK_TOTAL_COUNT( recent_counter_, NSMaxRange( operations_range_ ) ) ];

   ADSequence* sequence_copy_ = [ sequence_ copy ];

   [ sequence_copy_ async ];

   [ self waitForStatus: kGHUnitWaitStatusSuccess timeout: 2.0 ];
}

-(void)testSequenceOfSequences
{
   [ self prepare ];

   __block ADTOperationCounter recent_counter_ = 0;

   const NSInteger sequence_count_ = 15;

   NSRange operations_range_ = NSMakeRange( 0, 15 );

   NSMutableArray* sequences_ = [ NSMutableArray arrayWithCapacity: sequence_count_ ];

   //{{0..14} {15..29} .. {210..224}}
   for ( NSInteger i_ = 0; i_ < sequence_count_; ++i_ )
   {
      NSArray* operations_ = [ NSArray arrayWithOperationsFromRange: operations_range_
                                                          doneBlock: ADT_CHECK_AND_INCREMENT_COUNT( recent_counter_ ) ];

      [ sequences_ addObject: [ ADSequence compositeWithOperations: operations_
                                                              name: [ NSString stringWithFormat: @"%d", i_ ]
                                                         doneBlock: nil ] ];

      operations_range_.location += operations_range_.length;
   }

   ADSequence* global_sequence_ = [ ADSequence compositeWithOperations: sequences_
                                                                  name: @"global"
                                                             doneBlock: ADT_CHECK_TOTAL_COUNT( recent_counter_, operations_range_.location ) ];

   [ global_sequence_ async ];

   [ self waitForStatus: kGHUnitWaitStatusSuccess timeout: 2.0 ];
}

-(void)testSequenceOfConcurrents
{
   [ self prepare ];

   __block ADTOperationCounter recent_counter_ = 0;

   NSRange operations_range_ = NSMakeRange( 0, 15 );

   //{0..14
   NSMutableArray* operations_ = [ NSMutableArray arrayWithOperationsFromRange: operations_range_
                                                                     doneBlock: ADT_CHECK_AND_INCREMENT_COUNT( recent_counter_ ) ];

   operations_range_.location += operations_range_.length;

   //add [29(0.1 sec delay)..15(0 sec delay)]
   {
      NSArray* concurrent_operations_ = [ [ NSArray arrayWithOperationsFromRange: operations_range_
                                                                       doneBlock: ADT_CHECK_AND_INCREMENT_COUNT( recent_counter_ )
                                                                   delayFunction: ADTDirectDelay() ] reverseOrder ];

      ADConcurrent* concurrent_ = [ ADConcurrent compositeWithOperations: concurrent_operations_
                                                                    name: @"concurrent"
                                                               doneBlock: nil ];

      [ operations_ addObject: concurrent_ ];
   }

   operations_range_.location += operations_range_.length;

   //add 30..45}
   {
      [ operations_ addOperationsFromRange: operations_range_
                                 doneBlock: ADT_CHECK_AND_INCREMENT_COUNT( recent_counter_ ) ];
   }

   ADSequence* global_sequence_ = [ ADSequence compositeWithOperations: operations_
                                                                  name: @"global"
                                                             doneBlock: ADT_CHECK_TOTAL_COUNT( recent_counter_, NSMaxRange( operations_range_ ) ) ];

   [ global_sequence_ async ];

   [ self waitForStatus: kGHUnitWaitStatusSuccess timeout: 2.0 ];
}

-(void)testSequenceCancel
{
   [ self prepare ];

   NSRange operations_range_ = NSMakeRange( 0, 15 );

   __block ADTOperationCounter recent_counter_ = 0;

   //{0..14}
   ADSequence* sequence_ = [ ADSequence compositeWithOperations: [ NSArray arrayWithOperationsFromRange: operations_range_
                                                                                              doneBlock: ADT_CHECK_CANCEL_AND_INCREMENT_COUNT( recent_counter_ )
                                                                                          delayFunction: ADTConstDelay( 0.1 ) ]
                                                           name: @"global"
                                                      doneBlock: ADT_CHECK_TOTAL_COUNT( recent_counter_, NSMaxRange( operations_range_ ) ) ];

   id< ADRequest > request_ = [ sequence_ async ];
   [ request_ cancel ];

   [ self waitForStatus: kGHUnitWaitStatusSuccess timeout: 2.0 ];
}

-(void)testSequenceDoneBlock
{
   [ self prepare ];

   NSRange operations_range_ = NSMakeRange( 0, 15 );

   NSString* inner_sequence_name_ = @"Sequence";
   NSString* inner_operation_name_ = @"Operation";
   NSString* inner_concurrent_name_ = @"Concurrent";

   NSArray* expected_order_ = @[ inner_sequence_name_, inner_operation_name_, inner_concurrent_name_ ];
   NSMutableArray* real_order_ = [ NSMutableArray array ];

   ADSequence* inner_sequence_ = [ ADSequence compositeWithOperations: [ NSArray arrayWithOperationsFromRange: operations_range_
                                                                                              doneBlock: nil
                                                                                          delayFunction: ADTConstDelay( 0.1 ) ]
                                                           name: inner_sequence_name_
                                                      doneBlock: ^( id< ADResult > result_ )
                                  {
                                     [ real_order_ addObject: inner_sequence_name_ ];
                                  }];

   ADOperation* operation_ = [ [ ADBlockOperation alloc ] initWithName: inner_operation_name_ worker: ^id( NSError** error_ )
                              {
                                 return @(5);
                              }];
   
   operation_.doneBlock = ^( id< ADResult > result_ )
   {
      [ real_order_ addObject: inner_operation_name_ ];
   };

   
   ADConcurrent* inner_concurrent_ =
   [ ADConcurrent compositeWithOperations: [ NSArray arrayWithOperationsFromRange: operations_range_
                                                                        doneBlock: nil
                                                                    delayFunction: ADTConstDelay( 0.1 ) ]
                                     name: inner_concurrent_name_
                                doneBlock: ^( id< ADResult > result_ )
    {
       [ real_order_ addObject: inner_concurrent_name_ ];
    } ];

   ADSequence* global_sequence_ = [ [ ADSequence alloc ] initWithName: nil
                                                           operations: @[ inner_sequence_, operation_, inner_concurrent_ ] ];

   id< ADRequest > request_ = [ global_sequence_ async ];
   [ request_ wait ];

   GHAssertTrue( [ expected_order_ isEqualToArray: real_order_ ], @"Compare expected order: %@ with real order: %@", expected_order_, real_order_ );
}

-(void)testErrorInSequence
{
   NSRange operations_range_ = NSMakeRange( 0, 15 );
   
   __block ADTOperationCounter operations_success_ = 0;
   __block ADTOperationCounter operations_failed_ = 0;
   NSMutableArray* operations_ = [ NSMutableArray arrayWithOperationsFromRange: operations_range_
                                                                     doneBlock: ADT_INCREMENT_SUCCESS_CANCELLED( operations_success_, operations_failed_ ) ];

   operations_range_.location += operations_range_.length;

   [ operations_ addObject: [ ADBlockOperation operationWithName: @"failed"
                                                errorDescription: @"Error description"
                                                       doneBlock: ADT_INCREMENT_SUCCESS_CANCELLED( operations_success_, operations_failed_ ) ] ];

   operations_range_.location += 1;
   operations_range_.length = 10;

   [ operations_ addOperationsFromRange: operations_range_
                              doneBlock: ADT_INCREMENT_SUCCESS_CANCELLED( operations_success_, operations_failed_ ) ];

   ADSequence* sequence_ = [ ADSequence compositeWithOperations: operations_
                                                           name: @"global"
                                                      doneBlock: ^( id< ADResult > result_ )
                            {
                               NSLog( @"Result: %@", result_ );
                            }];

   id< ADRequest > request_ = [ sequence_ async ];
   [ request_ wait ];

   GHAssertTrue( operations_success_ == 15, @"Check success operations count" );
   GHAssertTrue( operations_failed_ == 10, @"Check failed operations count" );
}

-(void)testErrorInSequenceWithConcurrent
{
   NSRange operations_range_ = NSMakeRange( 0, 15 );
   
   __block ADTOperationCounter operations_success_ = 0;
   __block ADTOperationCounter operations_cancelled_ = 0;
   NSMutableArray* operations_ = [ NSMutableArray arrayWithOperationsFromRange: operations_range_
                                                                     doneBlock: ADT_INCREMENT_SUCCESS_CANCELLED( operations_success_, operations_cancelled_ ) ];
   
   operations_range_.location += operations_range_.length;
   
   [ operations_ addObject: [ ADBlockOperation operationWithName: @"failed"
                                                errorDescription: @"Error description"
                                                       doneBlock: ADT_INCREMENT_SUCCESS_CANCELLED( operations_success_, operations_cancelled_ ) ] ];
   
   operations_range_.location += 1;
   operations_range_.length = 10;
   
   {
      NSArray* concurrent_operations_ = [ NSArray arrayWithOperationsFromRange: operations_range_
                                                                     doneBlock: ADT_INCREMENT_SUCCESS_CANCELLED( operations_success_, operations_cancelled_ ) ];

      ADConcurrent* concurrent_ = [ ADConcurrent compositeWithOperations: concurrent_operations_
                                                                    name: @"concurrent"
                                                               doneBlock: nil ];

      [ operations_ addObject: concurrent_ ];
   }

   ADSequence* sequence_ = [ ADSequence compositeWithOperations: operations_
                                                           name: @"global"
                                                      doneBlock: ^( id< ADResult > result_ )
                            {
                               NSLog( @"Result: %@", result_ );
                            }];
   
   id< ADRequest > request_ = [ sequence_ async ];
   [ request_ wait ];

   NSLog(@"operations_success_: %d, operations_cancelled_: %d", operations_success_, operations_cancelled_);
   GHAssertTrue( operations_success_ == 15, @"Check success operations count" );
   GHAssertTrue( operations_cancelled_ == 10, @"Check cancelled operations count" );
}

@end
